#ifndef _PIN_HPP
#define _PIN_HPP
#pragma once

#include <avr/io.h>
#include "avr-defines.hpp"
#include "Reg.hpp"

/** expands PINDEF(B, 0) to &DDRB, &PORTB, &PINB, PB0
 * example: 
 *      OutputPin ledPin ( PINDEF(B, 0) );
 *   expands to: 
 *      OutputPin ledPin (&DDRB, &PORTB, &PINB, PB0);
 */
#define PINDEF(X,Y)  &DDR##X,&PORT##X,&PIN##X,Y

class Pin { /** base class for a pin */
protected:
     Register8 ddr;
     Register8 port;
     Register8 pin;
     uint8_t   num;

public:
    /** constructor 
     * @param  ddr   DDRn register address of pin (eg. &DDRB)
     * @param  port  PORTn register ''            (eg. &PORTB)
     * @param  pin   PINn register                (eg. &PINB)
     * @param  num   pin number                   (eg. PB2)
     */
     Pin(Reg8 ddr, Reg8 port, Reg8 pin, uint8_t num):
	ddr(ddr), port(port), pin(pin), num(num)
    {}
};


class InputPin : public Pin { /** an input pin */
public:
    /** constructor
     * @params  see Pin::Pin()
     * if compiled with "WITH_PULLUP" defined, add fifth bool 
     *   param for the pullup (generates larger code)
     */
    InputPin(Reg8 ddr,
	     Reg8 port, 
	     Reg8 pin, 
	     uint8_t num
#ifdef WITH_PULLUP
	     , const bool pullup=false
#endif //  WITH_PULLUP
	     ):
	Pin(ddr, port, pin, num) {
	this->ddr.cbi(this->num);
#ifdef WITH_PULLUP
	this->setPullup(pullup);
#endif // WITH_PULLUP
    }
    
    /** return true if pin is high, else false */
    bool isHigh(void) { return this->pin.gbi(this->num); };
    
    /** return true if pin is low, else false */
    bool isLow(void) { return !this->pin.gbi(this->num); };
    
    /** return true if pullup is set, else false */
    bool getPullup(void) { return this->port.gbi(this->num); }
    
    /** set pullup resistor 
     * @params  pullup  activate pullup if true
     */
    void setPullup(const bool pullup) { 
	if(pullup) { this->port.sbi(this->num);} 
	else { this->port.cbi(this->num); }
    }
};


class OutputPin : public Pin { /** an output pin */
public:
    OutputPin(Reg8 ddr,
	      Reg8 port, 
	      Reg8 pin, 
	      uint8_t num):
	/** constructor @params  see Pin::Pin() */
        Pin(ddr, port, pin, num) 
    { this->ddr.sbi(this->num); }

    /** set output high */
    void setHigh(void) { this->port.sbi(this->num); }

    /** set output low */
    void setLow(void)  { this->port.cbi(this->num);}

    /** toggle output */
    void toggle(void) {
#if CAN_TOGGLE_PIN == 1
	this->pin.sbi(this->num); 
#else
	if(this->pin.gbi(this->num)) {
	    this->setLow();
	} else { 
	    this->setHigh(); 
	}
#endif // CAN_TOGGLE_PIN
    }
    /** assignment operator 
     * @param  parm  if true: set pin high, else set pin low
     */
    void operator=(bool &parm) {
        if(parm) {
            this->setHigh();
        } else {
            this->setLow();
        }
    }
};


/**
 * deriving this from InputPin and OutputPin by virtually
 * inheriting from Pin bloats code so we better duplicate
 *
 * @todo think!
 */
class InoutPin : public Pin {
public:
    /** constructor
     * @params  see Pin::Pin()
     */
    InoutPin(Reg8 ddr, Reg8 port, Reg8 pin, uint8_t num):
	Pin(ddr, port, pin, num)
    {}

    /** set pin as input pin */
    void setInput(void) { this->ddr.cbi(this->num); }

    /** set pin as output pin */
    void setOutput(void) { this->ddr.sbi(this->num); }

    // input functions (see InputPin)
    int isHigh(void) { return this->pin.gbi(this->num); }
    int isLow(void) { return !this->pin.gbi(this->num); }

    // output functions (see OutputPin)
    void setHigh(void) { this->port.sbi(this->num); };
    void setLow(void)  { this->port.cbi(this->num); };
    void operator=(bool const &parm) {
        if(parm) {
            this->setHigh();
        } else {
            this->setLow();
        }
    }
};

#endif // _PIN_HPP
