#ifndef _REG_HPP
#define _REG_HPP
#pragma once

#include <inttypes.h>
#include <stdarg.h>

typedef  volatile uint8_t* const   Reg8;
typedef  volatile uint16_t* const  Reg16;

/** extract high and low uint8_t from a uint16_t */
#define LO_BYTE(byte) (uint8_t)((byte) & 0xFF)
#define HI_BYTE(byte) (uint8_t)(((byte) >> 8) & 0xFF)

/**
 * Register template class for 8- and 16-bit registers
 * @params  T    Reg8 or Reg16
 * @params  U    uint8_t or uint16_t
 * @params  reg  Register address (eg. &ADCSRA)
 */
template <typename T, typename U>
class Register {
protected:
    const T reg;
public:
    constexpr Register(T reg): reg(reg) {}
    /** set "bit" = 1 */
    constexpr void sbi(const U bit)  { *this->reg |= (1 << bit); }
    /** clear (unset, set to 0) "bit"  */
    constexpr void cbi(const U bit)  { *this->reg &= ~(1 << bit); }
    /** return true if "bit" is set */
    constexpr bool gbi(const U bit)  { return *this->reg & (1 << bit); }
    /** set bitmask "mask" */
    constexpr void sms(const U mask) { *this->reg |= mask; }
    /** clear bitmask "mask" */
    constexpr void cms(const U mask) { *this->reg &= ~mask;}
    /** set all "bits" */
    constexpr void set(const U bits) { *this->reg = bits;}
    /** get register value */
    constexpr U    get(void)   { return (U)*this->reg;  }

    /** EXPENSIVE! set bit list, eg. Register(&ADCSRA).sbiList(ADPS1, ADPS0)
     * @params  ...  list of bits to set
     */
    constexpr void sbl(uint8_t n, ...) {
	uint8_t val;
	va_list vl;
	va_start(vl, n);
	for(uint8_t i = 0; i < n; i++) {
	    val = va_arg(vl, int);
	    this->sbi(val);
	}
    }
    /** EXPENSIVE! clear bit list, eg. Register(&ADCSRA).sbiList(ADPS1, ADPS0)
     * @params  ...  list of bits to clear
     */
    constexpr void cbl(uint8_t n, ...) {
	uint8_t val;
	va_list vl;
	va_start(vl, n);
	for(uint8_t i = 0; i < n; i++) {
	    val = va_arg(vl, int);
	    this->cbi(val);
	}
    }
};

typedef Register< Reg8,  uint8_t  > Register8;
typedef Register< Reg16, uint16_t > Register16;

#endif //_REG_HPP
